import boto3
import json
import logging
import requests
import requests.exceptions


logger = logging.getLogger(__name__)


def fetch_json(uri):
    """ Perform an HTTP GET on the given uri, return the results as json.

    Args:
        uri: the string URI to fetch.

    Returns:
        A JSON object with the response or None if the status code of the
        response is an error code.
    """
    try:
        r = requests.get(uri)
        if r.status_code != requests.codes.ok:
            return None
        return r.json()
    except requests.exceptions.ConnectionError:
        return None


def get_s3_json_content(s3_bucket, s3_key):
    """Download and parse a json file stored on AWS S3.

    The file is downloaded and then cached for future use.
    """

    raw_data = None
    try:
        s3 = boto3.resource('s3')
        raw_data = (
            s3
            .Object(s3_bucket, s3_key)
            .get()['Body']
            .read()
            .decode('utf-8')
        )
    except Exception:
        logger.exception("Failed to download from S3", extra={
            "bucket": s3_bucket,
            "key": s3_key})
        return None

    # It can happen to have corrupted files. Account for the
    # sad reality of life.
    try:
        return json.loads(raw_data)
    except ValueError:
        logging.error("Cannot parse JSON resource from S3", extra={
            "bucket": s3_bucket,
            "key": s3_key})

    return None



import logging
import numpy as np
from .base_recommender import AbstractRecommender
from itertools import groupby
from scipy.spatial import distance

FLOOR_DISTANCE_ADJUSTMENT = 0.001

CATEGORICAL_FEATURES = ["geo_city", "locale", "os"]
CONTINUOUS_FEATURES = ["subsession_length", "bookmark_count", "tab_open_count", "total_uri", "unique_tlds"]

S3_BUCKET = 'telemetry-parquet'
DONOR_LIST_KEY = 'taar/similarity/donors.json'
LR_CURVES_SIMILARITY_TO_PROBABILITY = 'taar/similarity/lr_curves.json'

logger = logging.getLogger(__name__)


class SimilarityRecommender(AbstractRecommender):
    """ A recommender class that returns top N addons based on the
    client similarity with a set of candidate addon donors.

    Several telemetry fields are used to compute pairwise similarity
    with the donors and similarities are converted into a likelihood
    ratio of being a good match versus not being a good match. These
    quantities are then used to rank specific addons for
    recommendation.

    This will load a json file containing updated list of addon donors
    updated periodically by a separate weekly process using
    Longitdudinal Telemetry data.

    This recommender may provide useful recommendations when
    collaborative_recommender may not work.
    """

    def __init__(self, ctx):
        self._ctx = ctx
        assert 'cache' in self._ctx

        self._init_from_ctx()

    def _init_from_ctx(self):
        # Download the addon donors list.
        cache = self._ctx['cache']
        self.donors_pool = cache.get_s3_json_content(S3_BUCKET, DONOR_LIST_KEY)
        if self.donors_pool is None:
            logger.error("Cannot download the donor list: {}".format(DONOR_LIST_KEY))

        # Download the probability mapping curves from similarity to likelihood of being a good donor.
        self.lr_curves = cache.get_s3_json_content(S3_BUCKET, LR_CURVES_SIMILARITY_TO_PROBABILITY)
        if self.lr_curves is None:
            logger.error("Cannot download the lr curves: {}".format(LR_CURVES_SIMILARITY_TO_PROBABILITY))
        self.build_features_caches()

    def build_features_caches(self):
        """This function build two feature cache matrices.

        One matrix is for the continuous features and the other is for
        the categorical features. This is needed to speed up the similarity
        recommendation process."""
        if self.donors_pool is None or self.lr_curves is None:
            return None

        self.num_donors = len(self.donors_pool)

        # Build a numpy matrix cache for the continuous features.
        self.continuous_features = np.zeros((self.num_donors, len(CONTINUOUS_FEATURES)))
        for idx, d in enumerate(self.donors_pool):
            features = [d.get(specified_key) for specified_key in CONTINUOUS_FEATURES]
            self.continuous_features[idx] = features

        # Build the cache for categorical features.
        self.categorical_features =\
            np.zeros((self.num_donors, len(CATEGORICAL_FEATURES)), dtype='object')
        for idx, d in enumerate(self.donors_pool):
            features = [d.get(specified_key) for specified_key in CATEGORICAL_FEATURES]
            self.categorical_features[idx] = np.array([features], dtype="object")

    def can_recommend(self, client_data, extra_data={}):
        # We can't recommend if we don't have our data files.
        if self.donors_pool is None or self.lr_curves is None:
            return False

        # Check that the client info contains a non-None value for each required
        # telemetry field.
        REQUIRED_FIELDS = CATEGORICAL_FEATURES + CONTINUOUS_FEATURES

        has_fields = all([client_data.get(f, None) is not None for f in REQUIRED_FIELDS])
        if not has_fields:
            # Can not add extra info because client_id may not be available.
            logger.error("Unusable client data encountered")
        return has_fields

    def get_lr(self, score):
        """Compute a :float: likelihood ratio from a provided similarity score when compared
        to two probability density functions which are computed and pre-loaded during init.

        The numerator indicates the probability density that a particular similarity score
        corresponds to a 'good' addon donor, i.e. a client that is similar in the sense of
        telemetry variables. The denominator indicates the probability density that a particular
        similarity score corresponds to a 'poor' addon donor

        :param score: A similarity score between a pair of objects.
        :returns: The approximate float likelihood ratio corresponding to provided score.
        """
        # Find the index of the closest value that was precomputed in lr_curves
        # This will significantly speed up |get_lr|.

        # The lr_curves_cache is a list of scalar distance
        # measurements
        lr_curves_cache = np.array([s[0] for s in self.lr_curves])

        # np.argmin produces the index to the part of the curve
        # where distance is the smallest to the score which we are
        # inspecting currently.
        idx = np.argmin(abs(score - lr_curves_cache))

        numer_val = self.lr_curves[idx][1][0]
        denum_val = self.lr_curves[idx][1][1]

        # Compute LR based on numerator and denominator values
        return float(numer_val) / float(denum_val)

    # # # CAUTION! # # #
    # Any changes to this function must be reflected in the corresponding ETL job.
    # https://github.com/mozilla/python_mozetl/blob/master/mozetl/taar/taar_similarity.py
    #
    def compute_clients_dist(self, client_data):
        client_categorical_feats = [client_data.get(specified_key) for specified_key in CATEGORICAL_FEATURES]
        client_continuous_feats = [client_data.get(specified_key) for specified_key in CONTINUOUS_FEATURES]

        # Compute the distances between the user and the cached continuous features.
        cont_features = distance.cdist(
            self.continuous_features, np.array([client_continuous_feats]), 'canberra')

        # Compute the distances between the user and the cached categorical features.
        cat_features = np.array(
            [[distance.hamming(x, client_categorical_feats)] for x in self.categorical_features])

        # See the "Note about cdist optimization" in README.md for why we only use cdist once.

        # Take the product of similarities to attain a univariate similarity score.
        # Note that the addition of 0.001 to the continuous features
        # sets a floor value to the distance in continuous similarity
        # scores.  There is no such floor value set for categorical
        # features so this adjustment prioritizes categorical
        # similarity over continous similarity
        return (cont_features + FLOOR_DISTANCE_ADJUSTMENT) * cat_features

    def get_similar_donors(self, client_data):
        """Computes a set of :float: similarity scores between a client and a set of candidate
        donors for which comparable variables have been measured.

        A custom similarity metric is defined in this function that combines the Hamming distance
        for categorical variables with the Canberra distance for continuous variables into a
        univariate similarity metric between the client and a set of candidate donors loaded during
        init.

        :param client_data: a client data payload including a subset fo telemetry fields.
        :return: the sorted approximate likelihood ratio (np.array) corresponding to the
                 internally computed similarity score and a list of indices that link
                 each LR score with the related donor in the |self.donors_pool|.
        """
        # Compute the distance between self and any comparable client.
        distances = self.compute_clients_dist(client_data)

        # Compute the LR based on precomputed distributions that relate the score
        # to a probability of providing good addon recommendations.

        lrs_from_scores =\
            np.array([self.get_lr(distances[i]) for i in range(self.num_donors)])

        # Sort the LR values (descending) and return the sorted values together with
        # the original indices.
        indices = (-lrs_from_scores).argsort()
        return lrs_from_scores[indices], indices

    def recommend(self, client_data, limit, extra_data={}):
        donor_set_ranking, indices = self.get_similar_donors(client_data)
        donor_log_lrs = np.log(donor_set_ranking)
        # 1.0 corresponds to a log likelihood ratio of 0 meaning that donors are equally
        # likely to be 'good'. A value > 0.0 is sufficient, but we like this to be high.
        if donor_log_lrs[0] < 0.1:
            logger.warning("Addons recommended with very low similarity score, perhaps donor set is unrepresentative",
                           extra={"maximum_similarity": donor_set_ranking[0]})

        # Retrieve the indices of the highest ranked donors and then append their
        # installed addons.
        index_lrs_iter = zip(indices[donor_log_lrs > 0.0], donor_log_lrs)
        recommendations = []
        for (index, lrs) in index_lrs_iter:
            for term in self.donors_pool[index]['active_addons']:
                candidate = (term, lrs)
                recommendations.append(candidate)
        # Sort recommendations on key (guid name)
        recommendations = sorted(recommendations, key=lambda x: x[0])
        recommendations_out = []
        # recommendations must be sorted for this to work.
        for guid_key, group in groupby(recommendations, key=lambda x: x[0]):
            recommendations_out.append((guid_key, sum(j for i, j in group)))
        # now re-sort on the basis of LLR.
        recommendations_out = sorted(recommendations_out, key=lambda x: -x[1])

        log_data = (client_data['client_id'],
                    str([r[0] for r in recommendations_out[:limit]]))
        logger.info("similarity_recommender_triggered, "
                    "client_id: [%s], guids: [%s]" % log_data)

        return recommendations_out[:limit]



from taar.context import Context
from taar.cache import JSONCache, Clock

from taar.recommenders.ensemble_recommender import WeightCache, EnsembleRecommender
from .mocks import MockRecommenderFactory

EXPECTED = {'collaborative': 1000,
            'similarity': 100,
            'locale': 10}


class Mocker:
    def get_s3_json_content(self, *args, **kwargs):
        return {'ensemble_weights': EXPECTED}


def test_weight_cache():   # noqa

    ctx = Context()
    ctx['utils'] = Mocker()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)

    wc = WeightCache(ctx.child())
    actual = wc.getWeights()
    assert EXPECTED == actual


def test_recommendations():
    ctx = Context()

    ctx['utils'] = Mocker()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)

    EXPECTED_RESULTS = [('ghi', 3430.0),
                        ('def', 3320.0),
                        ('ijk', 3200.0),
                        ('hij', 3100.0),
                        ('lmn', 420.0)]

    factory = MockRecommenderFactory()
    ctx['recommender_factory'] = factory

    ctx['recommender_map'] = {'collaborative': factory.create('collaborative'),
                              'similarity': factory.create('similarity'),
                              'locale': factory.create('locale')}
    r = EnsembleRecommender(ctx.child())
    client = {'client_id': '12345'}  # Anything will work here

    recommendation_list = r.recommend(client, 5)
    assert isinstance(recommendation_list, list)
    assert recommendation_list == EXPECTED_RESULTS


def test_preinstalled_guids():
    ctx = Context()

    ctx['utils'] = Mocker()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)

    EXPECTED_RESULTS = [('ghi', 3430.0),
                        ('ijk', 3200.0),
                        ('lmn', 420.0),
                        ('klm', 409.99999999999994),
                        ('abc', 23.0)]

    factory = MockRecommenderFactory()
    ctx['recommender_factory'] = factory

    ctx['recommender_map'] = {'collaborative': factory.create('collaborative'),
                              'similarity': factory.create('similarity'),
                              'locale': factory.create('locale')}
    r = EnsembleRecommender(ctx.child())

    # 'hij' should be excluded from the suggestions list
    # The other two addon GUIDs 'def' and 'jkl' will never be
    # recommended anyway and should have no impact on results
    client = {'client_id': '12345',
              'installed_addons': ['def', 'hij', 'jkl']}

    recommendation_list = r.recommend(client, 5)
    print(recommendation_list)
    assert isinstance(recommendation_list, list)
    assert recommendation_list == EXPECTED_RESULTS



from taar.context import Context
from taar.cache import JSONCache, Clock

from taar.recommenders import LegacyRecommender

FAKE_LEGACY_DATA = {
    "guid-01":
        ["guid-02", 'guid-11-4'],
    "guid-03":
        ["guid-05", 'guid-16-4'],
    "guid-05":
        ["guid-06", 'guid-20-8'],
    "guid-07": ["guid-08-1", "guid-09-2", "guid-10-3"],
    "guid-12": ["guid-13-1", "guid-14-2", "guid-15-3",
                "guid-17-5", "guid-18-6", "guid-19-7",
                "guid-21-9", "guid-22-10"],
    "guid-23": []
}

LIMIT = 10


def s3_mocker(ctx):
    ctx = ctx.child()

    class Mocker:
        def get_s3_json_content(self, *args, **kwargs):
            return FAKE_LEGACY_DATA
    ctx['utils'] = Mocker()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    return ctx


def test_can_recommend():
    ctx = Context()
    ctx = s3_mocker(ctx)
    r = LegacyRecommender(ctx)

    # Test that we can't recommend if we have not enough client info.
    assert not r.can_recommend({})
    assert not r.can_recommend({"disabled_addons_ids": []})

    # Check that we can not recommend if no *legacy* addons are detected,
    # but addon is in loaded resource.
    profile_without_legacy = dict(
        client_id="test-client-id",
        disabled_addons_ids=["test_guid_7",
                             "test_guid_8"],
        locale="it-IT"
    )

    assert not r.can_recommend(profile_without_legacy)


def test_recommendations():
    """Test that the legacy recommender returns the correct addons from the json loaded.

    The JSON output for this recommender should be a list of 2-tuples
    of (GUID, weight).
    """
    ctx = Context()
    ctx = s3_mocker(ctx)
    r = LegacyRecommender(ctx)

    profile_with_many_legacy = dict(
        client_id="test-client-id",
        disabled_addons_ids=["guid-01",
                             "guid-05",
                             "guid-12"],
        locale="it-IT"
    )

    recommendations = r.recommend(profile_with_many_legacy, LIMIT)

    assert len(recommendations) == LIMIT
    assert ("guid-13-1", 1) in recommendations
    assert ("guid-21-9", 1) not in recommendations
    assert ("guid-22-10", 1) not in recommendations
    assert ("guid-21-9", 1) not in recommendations


def test_recommender_str():
    """Tests that the string representation of the recommender is correct
    """
    ctx = Context()
    ctx = s3_mocker(ctx)
    r = LegacyRecommender(ctx)
    assert str(r) == "LegacyRecommender"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from taar.context import Context
from taar.cache import JSONCache, Clock

from taar.profile_fetcher import ProfileFetcher
from taar.recommenders import RecommendationManager
from taar.recommenders.base_recommender import AbstractRecommender
from .test_ensemblerecommender import Mocker
from .mocks import MockProfileController, MockRecommenderFactory


class StubRecommender(AbstractRecommender):
    """ A shared, stub recommender that can be used for testing.
    """
    def __init__(self, can_recommend, stub_recommendations):
        self._can_recommend = can_recommend
        self._recommendations = stub_recommendations

    def can_recommend(self, client_info, extra_data={}):
        return self._can_recommend

    def recommend(self, client_data, limit, extra_data={}):
        return self._recommendations


def get_test_ctx():
    fetcher = ProfileFetcher(MockProfileController(None))
    factory = MockRecommenderFactory()
    ctx = Context()
    ctx['profile_fetcher'] = fetcher
    ctx['recommender_factory'] = factory

    # Just populate the utils key for test when WeightCache is
    # instantiated
    ctx['utils'] = None
    return ctx.child()


def test_none_profile_returns_empty_list():
    ctx = get_test_ctx()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    rec_manager = RecommendationManager(ctx)
    assert rec_manager.recommend("random-client-id", 10) == []


def test_recommendation_strategy():
    """The recommendation manager is currently very naive and just
    selects the first recommender which returns 'True' to
    can_recommend()."""

    EXPECTED_ADDONS = ["expected_id", "other-id"]

    # Create a stub ProfileFetcher that always returns the same
    # client data.
    class StubFetcher:
        def get(self, client_id):
            return {'client_id': '00000'}

    # Configure the recommender so that only the second model
    # can recommend and return the expected addons.
    factory = MockRecommenderFactory(legacy=lambda: StubRecommender(False, []),
                                     collaborative=lambda: StubRecommender(True, EXPECTED_ADDONS),
                                     similarity=lambda: StubRecommender(False, []),
                                     locale=lambda: StubRecommender(False, []))

    # Make sure the recommender returns the expected addons.
    ctx = get_test_ctx()
    ctx['recommender_factory'] = factory
    ctx['profile_fetcher'] = StubFetcher()
    ctx['utils'] = Mocker()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    manager = RecommendationManager(ctx.child())
    results = manager.recommend("client-id",
                                10,
                                extra_data={'branch': 'linear'})
    assert results == EXPECTED_ADDONS


def test_recommendations_via_manager():  # noqa
    ctx = get_test_ctx()

    EXPECTED_RESULTS = [('ghi', 3430.0),
                        ('def', 3320.0),
                        ('ijk', 3200.0),
                        ('hij', 3100.0),
                        ('lmn', 420.0),
                        ('klm', 409.99999999999994),
                        ('jkl', 400.0),
                        ('abc', 23.0),
                        ('fgh', 22.0),
                        ('efg', 21.0)]

    factory = MockRecommenderFactory()

    class MockProfileFetcher:
        def get(self, client_id):
            return {'client_id': client_id}

    ctx['recommender_factory'] = factory
    ctx['profile_fetcher'] = MockProfileFetcher()
    ctx['utils'] = Mocker()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    manager = RecommendationManager(ctx.child())
    recommendation_list = manager.recommend({'client_id': 'some_ignored_id'},
                                            10,
                                            extra_data={'branch': 'ensemble'})
    assert isinstance(recommendation_list, list)
    assert recommendation_list == EXPECTED_RESULTS



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from decouple import config
from flask import request
import json

# TAAR specific libraries
from taar.context import default_context
from taar.profile_fetcher import ProfileFetcher
from taar import recommenders
from taar import ProfileController

# These are configurations that are specific to the TAAR library
VALID_BRANCHES = set(['linear', 'ensemble', 'control'])
DYNAMO_REGION = config('DYNAMO_REGION', default='us-west-2')
DYNAMO_TABLE_NAME = config('DYNAMO_TABLE_NAME', default='taar_addon_data_20180206')
TAAR_MAX_RESULTS = config('TAAR_MAX_RESULTS', default=10, cast=int)


class ResourceProxy(object):
    def __init__(self):
        self._resource = None

    def setResource(self, rsrc):
        self._resource = rsrc

    def getResource(self):
        return self._resource


PROXY_MANAGER = ResourceProxy()


def configure_plugin(app):
    """
    This is a factory function that configures all the routes for
    flask given a particular library.
    """
    @app.route('/api/recommendations/<uuid:uuid_client_id>/')
    def recommendations(uuid_client_id):
        """Return a list of recommendations provided a telemetry client_id."""
        # Use the module global PROXY_MANAGER
        global PROXY_MANAGER

        # Coerce the uuid.UUID type into a string
        client_id = str(uuid_client_id)

        branch = request.args.get('branch', '')

        if branch.endswith('-taar'):
            branch = branch.replace("-taar", "")

        if branch not in VALID_BRANCHES:
            # Force branch to be a control branch if an invalid request
            # comes in.
            branch = 'control'

        extra_data = {'branch': branch}

        locale = request.args.get('locale', None)
        if locale is not None:
            extra_data['locale'] = locale

        platform = request.args.get('platform', None)
        if platform is not None:
            extra_data['platform'] = platform

        if PROXY_MANAGER.getResource() is None:
            ctx = default_context()
            dynamo_client = ProfileController(region_name=DYNAMO_REGION,
                                              table_name=DYNAMO_TABLE_NAME)
            profile_fetcher = ProfileFetcher(dynamo_client)

            ctx['profile_fetcher'] = profile_fetcher

            # Lock the context down after we've got basic bits installed
            root_ctx = ctx.child()
            r_factory = recommenders.RecommenderFactory(root_ctx)
            root_ctx['recommender_factory'] = r_factory
            instance = recommenders.RecommendationManager(root_ctx.child())
            PROXY_MANAGER.setResource(instance)

        instance = PROXY_MANAGER.getResource()
        recommendations = instance.recommend(client_id=client_id,
                                             limit=TAAR_MAX_RESULTS,
                                             extra_data=extra_data)
        # Strip out weights from TAAR results to maintain compatibility
        # with TAAR 1.0
        jdata = {"results": [x[0] for x in recommendations]}

        response = app.response_class(
                response=json.dumps(jdata),
                status=200,
                mimetype='application/json'
                )
        return response

    class MyPlugin:
        def set(self, config_options):
            """
            This setter is primarily so that we can instrument the
            cached RecommendationManager implementation under test.

            All plugins should implement this set method to enable
            overwriting configuration options with a TAAR library.
            """
            global PROXY_MANAGER
            if 'PROXY_RESOURCE' in config_options:
                PROXY_MANAGER._resource = config_options['PROXY_RESOURCE']

    return MyPlugin()




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import time
import threading


class Clock:
    def time(self):
        """Return epoch time in seconds like time.time()"""
        return time.time()


class JSONCache:
    """
    This class keeps a cache of JSON blobs and S3 bucket data.

    All data is expired simultaneously
    """
    def __init__(self, ctx):
        assert 'utils' in ctx
        assert 'clock' in ctx
        self._ctx = ctx

        # Set to 4 hours
        self._ttl = 60 * 60 * 4

        self._json_cache = {}
        self._s3_json_cache = {}

        self.refresh_expiry()

        self._lock = threading.RLock()

    def refresh_expiry(self):
        self._expiry_time = self._ctx['clock'].time() + self._ttl

    def fetch_json(self, url):
        with self._lock:
            utils = self._ctx['utils']
            if url not in self._json_cache:
                self._json_cache[url] = utils.fetch_json(url)
            content = self._json_cache[url]
            self.expire_cache()
            return content

    def get_s3_json_content(self, s3_bucket, s3_key):
        with self._lock:
            utils = self._ctx['utils']
            key = (s3_bucket, s3_key)
            if key not in self._s3_json_cache:
                self._s3_json_cache[key] = utils.get_s3_json_content(s3_bucket, s3_key)
            content = self._s3_json_cache[key]
            self.expire_cache()
            return content

    def expire_cache(self):
        with self._lock:
            clock = self._ctx['clock']

            if clock.time() >= self._expiry_time:
                self._json_cache.clear()
                self._s3_json_cache.clear()
                self.refresh_expiry()



from taar.recommenders import utils


def test_fetch_json():
    """ Just test a URL that we know will fail """
    jdata = utils.fetch_json("http://127.0.0.1:9001/some-nonexistant-url-foo.json")
    assert jdata is None


def test_get_s3_json_content():
    """ Just test an S3 bucket and key that doesn't exist """
    jdata = utils.get_s3_json_content("taar_not_my_bucket", "this/is/not/a/valid/path")
    assert jdata is None



from .collaborative_recommender import CollaborativeRecommender
from .locale_recommender import LocaleRecommender
from .legacy_recommender import LegacyRecommender
from .similarity_recommender import SimilarityRecommender
from .recommendation_manager import RecommendationManager, RecommenderFactory


__all__ = [
    'CollaborativeRecommender',
    'LegacyRecommender',
    'LocaleRecommender',
    'SimilarityRecommender',
    'RecommendationManager',
    'RecommenderFactory',
]



import logging
import numpy as np
import operator as op

from .base_recommender import AbstractRecommender

ADDON_MODEL_URL =\
    "https://s3-us-west-2.amazonaws.com/telemetry-public-analysis-2/telemetry-ml/addon_recommender/item_matrix.json"
ADDON_MAPPING_URL =\
    "https://s3-us-west-2.amazonaws.com/telemetry-public-analysis-2/telemetry-ml/addon_recommender/addon_mapping.json"

logger = logging.getLogger(__name__)


# http://garage.pimentech.net/libcommonPython_src_python_libcommon_javastringhashcode/
def java_string_hashcode(s):
    h = 0
    for c in s:
        h = (31 * h + ord(c)) & 0xFFFFFFFF
    return ((h + 0x80000000) & 0xFFFFFFFF) - 0x80000000


def positive_hash(s):
    return java_string_hashcode(s) & 0x7FFFFF


class CollaborativeRecommender(AbstractRecommender):
    """ The addon recommendation interface to the collaborative filtering model.

    Usage example::

        recommender = CollaborativeRecommender()
        dists = recommender.recommend(client_info)
    """
    def __init__(self, ctx):
        self._ctx = ctx

        assert 'cache' in self._ctx

        self._load_json_models()
        self.model = None
        self._build_model()

    def _load_json_models(self):
        # Download the addon mappings.
        self.addon_mapping = self._ctx['cache'].fetch_json(ADDON_MAPPING_URL)
        if self.addon_mapping is None:
            logger.error("Cannot download the addon mapping file {}".format(ADDON_MAPPING_URL))

        self.raw_item_matrix = self._ctx['cache'].fetch_json(ADDON_MODEL_URL)
        if self.addon_mapping is None:
            logger.error("Cannot download the model file {}".format(ADDON_MODEL_URL))

    def _build_model(self):
        if self.raw_item_matrix is None:
            return

        # Build a dense numpy matrix out of it.
        num_rows = len(self.raw_item_matrix)
        num_cols = len(self.raw_item_matrix[0]['features'])

        self.model = np.zeros(shape=(num_rows, num_cols))
        for index, row in enumerate(self.raw_item_matrix):
            self.model[index, :] = row['features']

    def can_recommend(self, client_data, extra_data={}):
        # We can't recommend if we don't have our data files.
        if self.raw_item_matrix is None or self.model is None or self.addon_mapping is None:
            return False

        # We only get meaningful recommendation if a client has at least an
        # addon installed.
        if len(client_data.get('installed_addons', [])) > 0:
            return True

        return False

    def recommend(self, client_data, limit, extra_data={}):
        # Addons identifiers are stored as positive hash values within the model.
        installed_addons_as_hashes =\
            [positive_hash(addon_id) for addon_id in client_data.get('installed_addons', [])]

        # Build the query vector by setting the position of the queried addons to 1.0
        # and the other to 0.0.
        query_vector = np.array([1.0
                                 if (entry.get("id") in installed_addons_as_hashes)
                                 else 0.0 for entry in self.raw_item_matrix])

        # Build the user factors matrix.
        user_factors = np.matmul(query_vector, self.model)
        user_factors_transposed = np.transpose(user_factors)

        # Compute the distance between the user and all the addons in the latent
        # space.
        distances = {}
        for addon in self.raw_item_matrix:
            # We don't really need to show the items we requested.
            # They will always end up with the greatest score. Also
            # filter out legacy addons from the suggestions.
            hashed_id = addon.get("id")
            str_hashed_id = str(hashed_id)
            if (hashed_id in installed_addons_as_hashes or
                    str_hashed_id not in self.addon_mapping or
                    self.addon_mapping[str_hashed_id].get("isWebextension", False) is False):
                continue

            dist = np.dot(user_factors_transposed, addon.get('features'))
            # Read the addon ids from the "addon_mapping" looking it
            # up by 'id' (which is an hashed value).
            addon_id = self.addon_mapping[str_hashed_id].get("id")
            distances[addon_id] = dist

        # Sort the suggested addons by their score and return the
        # sorted list of addon ids.
        sorted_dists = sorted(distances.items(),
                              key=op.itemgetter(1),
                              reverse=True)
        recommendations = [(s[0], s[1]) for s in sorted_dists[:limit]]

        log_data = (client_data['client_id'],
                    str([r[0] for r in recommendations]))
        logger.info("collaborative_recommender_triggered, "
                    "client_id: [%s], "
                    "guids: [%s]" % log_data)

        return recommendations



from taar.cache import Clock, JSONCache
import time
from taar.context import Context


EXPECTED_JSON = {"foo": 42}
EXPECTED_S3_JSON = {"foo": "bar"}


class MockUtils:
    def __init__(self):
        self._fetch_count = 0
        self._get_count = 0

    def fetch_json(self, url):
        self._fetch_count += 1
        return EXPECTED_JSON

    def get_s3_json_content(self, s3_bucket, s3_key):
        self._get_count += 1
        return EXPECTED_S3_JSON


def test_clock():
    cl = Clock()
    actual = cl.time()
    expected = time.time()

    # The clock should be pretty accurate to now
    assert abs(actual - expected) < 0.1


def test_fetch_json():
    """ Just test a URL that we know will fail """
    ctx = Context()
    ctx['utils'] = utils = MockUtils()
    ctx['clock'] = Clock()
    cache = JSONCache(ctx)
    jdata = cache.fetch_json("http://127.0.0.1:9001/some-nonexistant-url-foo.json")
    assert jdata == EXPECTED_JSON

    assert utils._fetch_count == 1
    for i in range(10):
        cache.fetch_json("http://127.0.0.1:9001/some-nonexistant-url-foo.json")
    assert utils._fetch_count == 1


def test_get_s3_json_content():
    """ Just test an S3 bucket and key that doesn't exist """
    ctx = Context()
    ctx['utils'] = utils = MockUtils()
    ctx['clock'] = Clock()
    cache = JSONCache(ctx)
    jdata = cache.get_s3_json_content("taar_not_my_bucket", "this/is/not/a/valid/path")
    assert jdata == EXPECTED_S3_JSON

    assert utils._get_count == 1
    for i in range(10):
        cache.get_s3_json_content("taar_not_my_bucket", "this/is/not/a/valid/path")
    assert utils._get_count == 1


def test_expiry():
    """ Just test a URL that we know will fail """
    class MockClock:
        def __init__(self):
            self._now = 100

        def time(self):
            return self._now

    ctx = Context()
    utils = MockUtils()
    ctx['utils'] = utils
    ctx['clock'] = MockClock()

    cache = JSONCache(ctx)

    cache._ttl = 0  # Set TTL to nothing
    cache.refresh_expiry()

    jdata = cache.fetch_json("http://127.0.0.1:9001/some-nonexistant-url-foo.json")
    assert jdata == EXPECTED_JSON
    jdata = cache.get_s3_json_content("taar_not_my_bucket", "this/is/not/a/valid/path")
    assert jdata == EXPECTED_S3_JSON

    assert utils._get_count == 1
    assert utils._fetch_count == 1

    for i in range(10):
        cache.fetch_json("http://127.0.0.1:9001/some-nonexistant-url-foo.json")
        cache.get_s3_json_content("taar_not_my_bucket", "this/is/not/a/valid/path")

    # Cache expires each time
    assert utils._get_count == 11
    assert utils._fetch_count == 11



import logging
from .base_recommender import AbstractRecommender

ADDON_LIST_BUCKET = 'telemetry-parquet'
ADDON_LIST_KEY = 'taar/locale/top10_dict.json'


logger = logging.getLogger(__name__)


class LocaleRecommender(AbstractRecommender):
    """ A recommender class that returns top N addons based on the client geo-locale.

    This will load a json file containing updated top n addons in use per geo locale
    updated periodically by a separate process on airflow using Longitdudinal Telemetry
    data.

    This recommender may provide useful recommendations when collaborative_recommender
    may not work.
    """
    def __init__(self, ctx):
        self._ctx = ctx
        assert 'cache' in self._ctx
        self._init_from_ctx()

    def _init_from_ctx(self):
        cache = self._ctx['cache']
        self.top_addons_per_locale = cache.get_s3_json_content(ADDON_LIST_BUCKET,
                                                               ADDON_LIST_KEY)
        if self.top_addons_per_locale is None:
            logger.error("Cannot download the top per locale file {}".format(ADDON_LIST_KEY))

    def can_recommend(self, client_data, extra_data={}):
        # We can't recommend if we don't have our data files.
        if self.top_addons_per_locale is None:
            return False

        # If we have data coming from other sources, we can use that for
        # recommending.
        client_locale = client_data.get('locale', None) or extra_data.get('locale', None)
        if not isinstance(client_locale, str):
            return False

        if client_locale not in self.top_addons_per_locale:
            return False

        if not self.top_addons_per_locale.get(client_locale):
            return False

        return True

    def recommend(self, client_data, limit, extra_data={}):
        # If we have data coming from multiple sourecs, prefer the one
        # from 'client_data'.
        client_locale = client_data.get('locale') or extra_data.get('locale', None)
        result_list = self.top_addons_per_locale.get(client_locale, [])[:limit]

        if 'locale' not in client_data:
            try:
                client_data['locale'] = extra_data["locale"]
            except KeyError:
                client_data['locale'] = None

        log_data = (client_data['locale'],
                    str([r for r in result_list]))
        logger.info("locale_recommender_triggered, "
                    "client_locale: [%s], guids: [%s]" % log_data)
        # Need to have the actual prevalence here in order to improve this.
        return [(x, 1.0) for x in result_list]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Context
"""
from taar.context import Context


def test_context():
    ctx = Context()
    ctx['foo'] = 42
    child_ctx = ctx.child()
    assert child_ctx['foo'] == 42

    # Now clobber the local context, and demonstrate
    # that we haven't touched the parent
    child_ctx['foo'] = 'bar'
    assert child_ctx['foo'] == 'bar'
    assert child_ctx.get('foo', 'batz') == 'bar'
    assert ctx['foo'] == 42
    assert ctx.get('foo', 'bar') == 42

    # Revert the child back to the parent value
    del child_ctx['foo']
    assert child_ctx['foo'] == 42

    # Defaults work as expected
    assert child_ctx.get('foo', 'bar') == 42



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import boto3
import json
import logging
import zlib

logger = logging.getLogger(__name__)


class ProfileController:
    """
    This class provides basic read/write access into a AWS DynamoDB
    backed datastore.  The profile controller and profile fetcher code
    should eventually be merged as individually they don't "pull their
    weight".
    """

    def __init__(self, region_name, table_name):
        """
        Configure access to the DynamoDB instance
        """
        self._ddb = boto3.resource('dynamodb', region_name=region_name)
        self._table = self._ddb.Table(table_name)

    def get_client_profile(self, client_id):
        """This fetches a single client record out of DynamoDB
        """
        try:
            response = self._table.get_item(Key={'client_id': client_id})
            compressed_bytes = response['Item']['json_payload'].value
            json_byte_data = zlib.decompress(compressed_bytes)
            json_str_data = json_byte_data.decode('utf8')
            return json.loads(json_str_data)
        except Exception:
            # Return None on error.  The caller in ProfileFetcher will
            # handle error logging
            logger.error("Error loading client data for [%s]" % client_id)
            return None



import pytest
from taar.context import default_context
from taar import ProfileController, ProfileFetcher
from taar import recommenders
import time


def create_recommendation_manager():
    root_ctx = default_context()
    client = ProfileController('us-west-2', 'taar_addon_data_20180206')
    pf = ProfileFetcher(client)
    root_ctx['profile_fetcher'] = pf
    r_factory = recommenders.RecommenderFactory(root_ctx.child())
    root_ctx['recommender_factory'] = r_factory
    rm = recommenders.RecommendationManager(root_ctx.child())
    return rm


@pytest.mark.skip("This is an integration test")
def test_recommenders(client_id='some_dev_client_id', branch='linear'):
    """
    This integration test can be used to drive the TAAR
    RecommendationManager from your machine.  This assumes you have
    access to the proper dev IAM roles to read the S3 buckets.

    Note that the RecommendationManager emits addon GUID and a weight.
    The taar-api webservice does not display the weights in the JSON
    output.

    You can drive taar in your virtualenv with ipython like this :

            Victors-MacBook-Pro-2:~ victorng$ cd dev/taar
            Victors-MacBook-Pro-2:taar victorng$ workon taar
            (taar) Victors-MacBook-Pro-2:taar victorng$ ipython
            Python 3.5.4 (default, Jan 11 2018, 09:58:51)
            Type 'copyright', 'credits' or 'license' for more information
            IPython 6.2.1 -- An enhanced Interactive Python. Type '?' for help.

            In [1]: client_id = '6d93e4e9-96ca-433d-a4a9-92efd0da8957'

            In [2]: from tests.test_integration import test_recommenders

            In [3]: test_recommenders(client_id=client_id, branch='ensemble')
            Out[3]:
            [('uBlock0@raymondhill.net', 2.11520519710274),
             ('YoutubeDownloader@PeterOlayev.com', 2.09866473),
             ('artur.dubovoy@gmail.com', 2.09866473),
             ('enhancerforyoutube@maximerf.addons.mozilla.org', 2.09866473),
             ('jid1-NIfFY2CA8fy1tg@jetpack', 2.09866473),
             ('translator@zoli.bod', 2.09866473),
             ('{b9db16a4-6edc-47ec-a1f4-b86292ed211d}', 2.09866473),
             ('{bee6eb20-01e0-ebd1-da83-080329fb9a3a}', 2.09866473),
             ('{d10d0bf8-f5b5-c8b4-a8b2-2b9879e08c5d}', 2.09866473),
             ('{e4a8a97b-f2ed-450b-b12d-ee082ba24781}', 2.09866473)]

            In [4]:
    """
    rm = create_recommendation_manager()
    result = rm.recommend(client_id, limit=10, extra_data={'branch': branch})
    return result


def micro_bench(x, client_id, branch_label):
    """
    Run a microbenchmark against the recommendation manager.

    This should really be run in an AWS enviroment so that you're not
    hitting as much network latency when running from your local
    laptop.
    """
    rm = create_recommendation_manager()
    start = time.time()
    for i in range(x):
        rm.recommend(client_id, limit=10, extra_data={'branch': branch_label})
    end = time.time()

    print(("%0.5f seconds per request" % ((end - start) / x)))



"""
Test cases for the TAAR CollaborativeRecommender
"""

import numpy

from taar.context import Context
from taar.cache import JSONCache, Clock

from taar.recommenders.collaborative_recommender import ADDON_MAPPING_URL
from taar.recommenders.collaborative_recommender import ADDON_MODEL_URL


from taar.recommenders.collaborative_recommender import CollaborativeRecommender
from taar.recommenders.collaborative_recommender import positive_hash

"""
We need to generate a synthetic list of addons and relative weights
for co-occurance.  It's important to note that the
CollaborativeRecommender model expects that addon IDs are hashed using
the Java hash function.
"""

ADDON_SPACE = [{"id": "addon1.id", "name": "addon1.name", "isWebextension": True},
               {"id": "addon2.id", "name": "addon2.name", "isWebextension": True},
               {"id": "addon3.id", "name": "addon3.name", "isWebextension": True},
               {"id": "addon4.id", "name": "addon4.name", "isWebextension": True},
               {"id": "addon5.id", "name": "addon5.name", "isWebextension": True}]

# Load the addons into the FAKE_MAPPING dictionary
FAKE_MAPPING = {}
for addon in ADDON_SPACE:
    java_hash = positive_hash(addon['id'])
    FAKE_MAPPING[str(java_hash)] = addon

# This matrix sets up addon2 as an overweighted recommended addon
FAKE_ADDON_MATRIX = []
for i, addon in enumerate(ADDON_SPACE):
    row = {"id": positive_hash(addon['id']), "features": [0, 0.2, 0.0, 0.1, 0.15]}
    row['features'][i] = 1.0
    FAKE_ADDON_MATRIX.append(row)

"""
FAKE_ADDON_MATRIX = [
    {"id": 1234, "features": [1.0, 0.0, 0.0]},
    {"id": 4567, "features": [0.0, 1.0, 0.0]},
    {"id": 7890, "features": [0.0, 0.0, 1.0]}
]
"""


def activate_error_responses(ctx):
    """
    Overload the 'real' addon model and mapping URLs responses so that
    we always get 404 errors.
    """
    ctx = ctx.child()

    class ErrorUtils:
        def fetch_json(self, url):
            return None
    ctx['utils'] = ErrorUtils()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    return ctx


def activate_responses(ctx):
    """
    Overload the 'real' addon model and mapping URLs responses so that
    we always the fixture data at the top of this test module.
    """
    # responses.add(responses.GET, ADDON_MODEL_URL, json=FAKE_ADDON_MATRIX)
    # responses.add(responses.GET, ADDON_MAPPING_URL, json=FAKE_MAPPING)
    ctx = ctx.child()

    class MockUtils:
        def fetch_json(self, url):
            if url == ADDON_MODEL_URL:
                return FAKE_ADDON_MATRIX
            elif url == ADDON_MAPPING_URL:
                return FAKE_MAPPING

    ctx['utils'] = MockUtils()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    return ctx


def test_can_recommend():
    ctx = get_mocked_ctx()
    r = CollaborativeRecommender(ctx)

    # Test that we can't recommend if we have not enough client info.
    assert not r.can_recommend({})
    assert not r.can_recommend({"installed_addons": []})

    # Check that we can recommend if we the user has at least an addon.
    assert r.can_recommend({"installed_addons": ["uBlock0@raymondhill.net"],
                            "client_id": "test-client"})


def get_error_ctx():
    ctx = Context()
    ctx = activate_error_responses(ctx)
    return ctx


def get_mocked_ctx():
    ctx = Context()
    ctx = activate_responses(ctx)
    return ctx


def test_can_recommend_no_model():
    ctx = get_error_ctx()
    r = CollaborativeRecommender(ctx)

    # We should never be able to recommend if something went wrong with the model.
    assert not r.can_recommend({})
    assert not r.can_recommend({"installed_addons": []})
    assert not r.can_recommend({"installed_addons": ["uBlock0@raymondhill.net"]})


def test_empty_recommendations():
    # Tests that the empty recommender always recommends an empty list
    # of addons if we have no addons
    ctx = get_mocked_ctx()
    r = CollaborativeRecommender(ctx)
    assert not r.can_recommend({})

    # Note that calling recommend() if can_recommend has failed is not
    # defined.


def test_best_recommendation():
    # Make sure the structure of the recommendations is correct and that we
    # recommended the the right addon.
    ctx = get_mocked_ctx()
    r = CollaborativeRecommender(ctx)

    # An non-empty set of addons should give a list of recommendations
    fixture_client_data = {"installed_addons": ["addon4.id"],
                           "client_id": "test_client"}
    assert r.can_recommend(fixture_client_data)
    recommendations = r.recommend(fixture_client_data, 1)

    assert isinstance(recommendations, list)
    assert len(recommendations) == 1

    # Verify that addon2 - the most heavy weighted addon was
    # recommended
    result = recommendations[0]
    assert type(result) is tuple
    assert len(result) == 2
    assert result[0] == 'addon2.id'
    assert type(result[1]) is numpy.float64
    assert numpy.isclose(result[1], numpy.float64('0.3225'))


def test_recommendation_weights():
    """
    Weights should be ordered greatest to lowest
    """
    ctx = get_mocked_ctx()
    r = CollaborativeRecommender(ctx)

    # An non-empty set of addons should give a list of recommendations
    fixture_client_data = {"installed_addons": ["addon4.id"],
                           "client_id": "test_client"}
    assert r.can_recommend(fixture_client_data)
    recommendations = r.recommend(fixture_client_data, 2)
    assert isinstance(recommendations, list)
    assert len(recommendations) == 2

    # Verify that addon2 - the most heavy weighted addon was
    # recommended
    result = recommendations[0]
    assert type(result) is tuple
    assert len(result) == 2
    assert result[0] == 'addon2.id'
    assert type(result[1]) is numpy.float64
    assert numpy.isclose(result[1], numpy.float64('0.3225'))

    # Verify that addon2 - the most heavy weighted addon was
    # recommended
    result = recommendations[1]
    assert type(result) is tuple
    assert len(result) == 2
    assert result[0] == 'addon5.id'
    assert type(result[1]) is numpy.float64
    assert numpy.isclose(result[1], numpy.float64('0.29'))


def test_recommender_str():
    """Tests that the string representation of the recommender is correct
    """
    # TODO: this test is brittle and should be removed once it is safe
    # to do so
    ctx = get_mocked_ctx()
    r = CollaborativeRecommender(ctx)
    assert str(r) == "CollaborativeRecommender"




from taar.adapters.dynamo import ProfileController
import boto3
import zlib
import json


def test_crashy_profile_controller(monkeypatch):
    def mock_boto3_resource(*args, **kwargs):
        class ExceptionRaisingMockTable:
            def __init__(self, tbl_name):
                pass

            def get_item(self, *args, **kwargs):
                raise Exception

        class MockDDB:
            pass
        mock_ddb = MockDDB()
        mock_ddb.Table = ExceptionRaisingMockTable
        return mock_ddb

    monkeypatch.setattr(boto3, 'resource', mock_boto3_resource)

    pc = ProfileController('us-west-2', 'taar_addon_data_20180206')
    assert pc.get_client_profile("exception_raising_client_id") is None


def test_profile_controller(monkeypatch):
    def mock_boto3_resource(*args, **kwargs):
        some_bytes = zlib.compress(json.dumps({'key': "with_some_data"}).encode('utf8'))

        class ValueObj:
            value = some_bytes

        class MockTable:
            def __init__(self, tbl_name):
                pass

            def get_item(self, *args, **kwargs):
                value_obj = ValueObj()
                response = {'Item': {'json_payload': value_obj}}
                return response

        class MockDDB:
            pass
        mock_ddb = MockDDB()
        mock_ddb.Table = MockTable
        return mock_ddb

    monkeypatch.setattr(boto3, 'resource', mock_boto3_resource)

    pc = ProfileController('us-west-2', 'taar_addon_data_20180206')
    jdata = pc.get_client_profile("exception_raising_client_id")
    assert jdata == {'key': 'with_some_data'}



import logging

logger = logging.getLogger(__name__)


class ProfileFetcher(object):
    """ Fetch the latest information for a client on the backing
    datastore
    """
    def __init__(self, client):
        self.client = client

    def get(self, client_id):
        profile_data = self.client.get_client_profile(client_id)

        if profile_data is None:
            logger.error("Client profile not found", extra={"client_id": client_id})
            return None

        addon_ids = [addon['addon_id']
                     for addon in profile_data.get('active_addons', [])
                     if not addon.get('is_system', False)]

        return {
            "client_id": client_id,
            "geo_city": profile_data.get("city", ''),
            "subsession_length": profile_data.get("subsession_length", 0),
            "locale": profile_data.get('locale', ''),
            "os": profile_data.get("os", ''),
            "installed_addons": addon_ids,
            "disabled_addons_ids": profile_data.get("disabled_addons_ids", []),
            "bookmark_count": profile_data.get("places_bookmarks_count", 0),
            "tab_open_count": profile_data.get("scalar_parent_browser_engagement_tab_open_event_count", 0),
            "total_uri": profile_data.get("scalar_parent_browser_engagement_total_uri_count", 0),
            "unique_tlds": profile_data.get("scalar_parent_browser_engagement_unique_domains_count", 0),
        }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


class MockRecommender:
    """The MockRecommender takes in a map of GUID->weight."""

    def __init__(self, guid_map):
        self._guid_map = guid_map

    def can_recommend(self, *args, **kwargs):
        return True

    def recommend(self, *args, **kwargs):
        return sorted(self._guid_map.items(), key=lambda item: -item[1])


class MockRecommenderFactory:
    """
    A RecommenderFactory provides support to create recommenders.

    The existence of a factory enables injection of dependencies into
    the RecommendationManager and eases the implementation of test
    harnesses.
    """
    def __init__(self, **kwargs):
        mock_legacy = MockRecommender({'abc': 1.0, 'bcd': 1.1, 'cde': 1.2})
        mock_locale = MockRecommender({'def': 2.0, 'efg': 2.1, 'fgh': 2.2, 'abc': 2.3})
        mock_collaborative = MockRecommender({'ghi': 3.0, 'hij': 3.1, 'ijk': 3.2, 'def': 3.3})
        mock_similarity = MockRecommender({'jkl': 4.0,  'klm': 4.1, 'lmn': 4.2, 'ghi': 4.3})

        self._recommender_factory_map = {'legacy': lambda: mock_legacy,
                                         'collaborative': lambda: mock_collaborative,
                                         'similarity': lambda: mock_similarity,
                                         'locale': lambda: mock_locale}

        # Clobber any kwarg passed in recommenders
        for key in self._recommender_factory_map.keys():
            self._recommender_factory_map[key] = kwargs.get(key, self._recommender_factory_map[key])

    def get_names(self):
        return self._recommender_factory_map.keys()

    def create(self, recommender_name):
        return self._recommender_factory_map[recommender_name]()


class MockProfileController:
    def __init__(self, mock_profile):
        self._profile = mock_profile

    def get_client_profile(self, client_id):
        return self._profile



import logging
from taar.recommenders.ensemble_recommender import EnsembleRecommender


logger = logging.getLogger(__name__)


class RecommenderFactory:
    """
    A RecommenderFactory provides support to create recommenders.

    The existence of a factory enables injection of dependencies into
    the RecommendationManager and eases the implementation of test
    harnesses.
    """
    def __init__(self, ctx):
        self._ctx = ctx
        self._recommender_factory_map = self._ctx['recommender_factory_map']

    def get_names(self):
        return self._recommender_factory_map.keys()

    def create(self, recommender_name):
        return self._recommender_factory_map[recommender_name]()


class RecommendationManager:
    """This class determines which of the set of recommendation
    engines will actually be used to generate recommendations."""

    LINEAR_RECOMMENDER_ORDER = ['legacy', 'collaborative', 'similarity', 'locale']

    def __init__(self, ctx):
        """Initialize the user profile fetcher and the recommenders.
        """
        self._ctx = ctx

        assert 'recommender_factory' in self._ctx
        assert 'profile_fetcher' in self._ctx

        recommender_factory = ctx['recommender_factory']
        profile_fetcher = ctx['profile_fetcher']

        self.profile_fetcher = profile_fetcher
        self.linear_recommenders = []
        self._recommender_map = {}

        logger.info("Initializing recommenders")
        for rkey in self.LINEAR_RECOMMENDER_ORDER:
            recommender = recommender_factory.create(rkey)

            self.linear_recommenders.append(recommender)
            self._recommender_map[rkey] = recommender

        # Install the recommender_map to the context and instantiate
        # the EnsembleRecommender
        self._ctx['recommender_map'] = self._recommender_map
        self._recommender_map['ensemble'] = EnsembleRecommender(self._ctx.child())

    def recommend(self, client_id, limit, extra_data={}):
        """Return recommendations for the given client.

        The recommendation logic will go through each recommender and
        pick the first one that "can_recommend".

        :param client_id: the client unique id.
        :param limit: the maximum number of recommendations to return.
        :param extra_data: a dictionary with extra client data.
        """
        # Get the info for the requested client id.
        client_info = self.profile_fetcher.get(client_id)
        if client_info is None:
            return []

        # Compute the recommendation.

        # Select recommendation output based on extra_data['branch']
        branch_selector = extra_data.get('branch', 'control')
        if branch_selector not in ('control', 'linear', 'ensemble'):
            return []
        branch_method = getattr(self, 'recommend_%s' % branch_selector)
        return branch_method(client_info, client_id, limit, extra_data)

    def recommend_control(self, client_info, client_id, limit, extra_data):
        """Run the control recommender - that is nothing"""
        return []

    def recommend_ensemble(self, client_info, client_id, limit, extra_data):
        """Run the ensemble recommender """
        recommender = self._recommender_map['ensemble']
        return recommender.recommend(client_info, limit, extra_data)

    def recommend_linear(self, client_info, client_id, limit, extra_data):
        """Run the linear recommender"""
        for r in self.linear_recommenders:
            if r.can_recommend(client_info, extra_data):
                logger.info("Recommender selected", extra={
                    "client_id": client_id, "recommender": str(r)
                })
                recommendations = r.recommend(client_info, limit, extra_data)
                if not recommendations:
                    logger.info("No recommendations", extra={
                        "client_id": client_id, "recommender": str(r)
                    })
                else:
                    logger.info("Recommendations served", extra={
                        "client_id": client_id, "recommended_addons": str(recommendations)
                    })
                return recommendations
        logger.info("No recommender can recommend addons", extra={
            "client_id": client_id
        })
        return []



from taar import ProfileFetcher


class MockProfileController:
    def __init__(self, mock_profile):
        self._profile = mock_profile

    def get_client_profile(self, client_id):
        return self._profile


def test_profile_fetcher_returns_none():
    fetcher = ProfileFetcher(MockProfileController(None))
    assert fetcher.get("random-client-id") is None


MOCK_DATA = {'profile': {u'scalar_parent_browser_engagement_total_uri_count': 791,
                         u'city': u'Rome',
                         u'scalar_parent_browser_engagement_tab_open_event_count': 46,
                         u'subsession_start_date': u'2017-09-20T10:00:00.0+02:00',
                         u'subsession_length': 3785,
                         u'places_bookmarks_count': 0,
                         u'scalar_parent_browser_engagement_unique_domains_count': 11,
                         u'os': u'Windows_NT',
                         u'active_addons': [{u'addon_id': u'e10srollout@mozilla.org'},
                                            {u'addon_id': u'firefox@getpocket.com'},
                                            {u'addon_id': u'webcompat@mozilla.org', 'is_system': True}],
                         u'locale': 'it-IT'},
             'expected_result': {"client_id": 'random-client-id',
                                 "bookmark_count": 0,
                                 "disabled_addons_ids": [],
                                 "geo_city": "Rome",
                                 "os": "Windows_NT",
                                 "subsession_length": 3785,
                                 "tab_open_count": 46,
                                 "total_uri": 791,
                                 "unique_tlds": 11,
                                 "installed_addons": ["e10srollout@mozilla.org",
                                                      "firefox@getpocket.com"],
                                 "locale": "it-IT"}
             }


def test_profile_fetcher_returns_dict():
    mock_data = MOCK_DATA['profile']
    mock_profile_controller = MockProfileController(mock_data)
    fetcher = ProfileFetcher(mock_profile_controller)

    # Note that active_addons in the raw JSON source is remapped to
    # 'installed_addons'
    assert fetcher.get("random-client-id") == MOCK_DATA['expected_result']


def test_dont_crash_without_active_addons():
    mock_data = MOCK_DATA['profile']
    del mock_data['active_addons']

    mock_profile_controller = MockProfileController(mock_data)
    fetcher = ProfileFetcher(mock_profile_controller)
    expected = MOCK_DATA['expected_result']
    expected['installed_addons'][:] = []
    assert fetcher.get("random-client-id") == expected



from taar.context import Context
from taar.cache import JSONCache, Clock

from taar.recommenders import LocaleRecommender


FAKE_LOCALE_DATA = {
    "te-ST": [
        "{1e6b8bce-7dc8-481c-9f19-123e41332b72}", "some-other@nice-addon.com",
        "{66d1eed2-a390-47cd-8215-016e9fa9cc55}", "{5f1594c3-0d4c-49dd-9182-4fbbb25131a7}"
    ],
    "en": [
        "some-uuid@test-addon.com", "other-addon@some-id.it"
    ]
}


class MockUtils:
    def get_s3_json_content(self, *args, **kwargs):
        return FAKE_LOCALE_DATA


def create_test_ctx():
    ctx = Context()
    ctx['utils'] = MockUtils()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    return ctx.child()


def test_can_recommend():
    ctx = create_test_ctx()
    r = LocaleRecommender(ctx)

    # Test that we can't recommend if we have not enough client info.
    assert not r.can_recommend({})
    assert not r.can_recommend({"locale": []})

    # Check that we can recommend if the user has at least an addon.
    assert r.can_recommend({"locale": "en"})


def test_can_recommend_no_model():
    ctx = create_test_ctx()
    r = LocaleRecommender(ctx)

    # We should never be able to recommend if something went
    # wrong with the model.
    assert not r.can_recommend({})
    assert not r.can_recommend({"locale": []})
    assert not r.can_recommend({"locale": "it"})


def test_recommendations():
    """Test that the locale recommender returns the correct
    locale dependent addons.

    The JSON output for this recommender should be a list of 2-tuples
    of (GUID, weight).
    """
    ctx = create_test_ctx()
    r = LocaleRecommender(ctx)
    recommendations = r.recommend({"locale": "en"}, 10)

    # Make sure the structure of the recommendations is correct and that we
    # recommended the the right addon.
    assert isinstance(recommendations, list)
    assert len(recommendations) == len(FAKE_LOCALE_DATA["en"])

    # Make sure that the reported addons are the one from the fake data.
    for (addon_id, weight) in recommendations:
        assert 1 == weight
        assert addon_id in FAKE_LOCALE_DATA["en"]


def test_recommender_str():
    """Tests that the string representation of the recommender is correct
    """
    # TODO: this test is brittle and should be removed once it is safe
    # to do so
    ctx = create_test_ctx()
    r = LocaleRecommender(ctx)
    assert str(r) == "LocaleRecommender"


def test_recommender_extra_data():
    # Test that the recommender uses locale data from the "extra"
    # section if available.
    def validate_recommendations(data, expected_locale):
        # Make sure the structure of the recommendations is correct and that we
        # recommended the the right addon.
        assert isinstance(data, list)
        assert len(data) == len(FAKE_LOCALE_DATA[expected_locale])

        # Make sure that the reported addons are the one from the fake data.
        for (addon_id, weight) in data:
            assert addon_id in FAKE_LOCALE_DATA[expected_locale]
            assert 1 == weight

    ctx = create_test_ctx()
    r = LocaleRecommender(ctx)
    recommendations = r.recommend({}, 10, extra_data={"locale": "en"})
    validate_recommendations(recommendations, "en")

    # Make sure that we favour client data over the extra data.
    recommendations = r.recommend({"locale": "en"}, 10, extra_data={"locale": "te-ST"})
    validate_recommendations(recommendations, "en")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import six

import numpy as np
import scipy.stats

from taar.context import Context
from taar.cache import JSONCache, Clock

from taar.recommenders.similarity_recommender import \
    CATEGORICAL_FEATURES, CONTINUOUS_FEATURES, DONOR_LIST_KEY, LR_CURVES_SIMILARITY_TO_PROBABILITY, \
    SimilarityRecommender

from .similarity_data import CONTINUOUS_FEATURE_FIXTURE_DATA
from .similarity_data import CATEGORICAL_FEATURE_FIXTURE_DATA


def generate_fake_lr_curves(num_elements, ceiling=10.0):
    """
    Generate a mock likelihood ratio (LR) curve that can be used for
    testing.
    """
    lr_index = list(np.linspace(0, ceiling, num_elements))

    # This sets up a normal distribution with a mean of 0.5 and std
    # deviation of 0.5
    numerator_density = [scipy.stats.norm.pdf(float(i), 0.5, 0.5) for i in lr_index]

    # This sets up a normal distribution with a mean of 5.0 and std
    # deviation of 1.5.  So this is right shifted and has a wide std
    # deviation compared to the first curve.

    # This results in a small overlap.  Using ranges between 0.5 to
    # 5.0 will yield large values to small values.
    denominator_density = [scipy.stats.norm.pdf(float(i), 5.0, 1.5) for i in lr_index]
    return list(zip(lr_index, zip(numerator_density, denominator_density)))


def generate_a_fake_taar_client():
    return {
        "client_id": "test-client-001",
        "activeAddons": [],
        "geo_city": "brasilia-br",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 222,
        "unique_tlds": 21
    }


class MockNoDataUtils:
    def get_s3_json_content(self, *args, **kwargs):
        return None


class MockCategoricalData:

    cat_data = json.loads(json.dumps(CATEGORICAL_FEATURE_FIXTURE_DATA))
    lrs_data = json.loads(json.dumps(generate_fake_lr_curves(1000)))

    def get_s3_json_content(self, bucket, key):
        if key == DONOR_LIST_KEY:
            return self.cat_data
        if key == LR_CURVES_SIMILARITY_TO_PROBABILITY:
            return self.lrs_data


class MockContinuousData:

    cts_data = json.loads(json.dumps(CONTINUOUS_FEATURE_FIXTURE_DATA))
    lrs_data = json.loads(json.dumps(generate_fake_lr_curves(1000)))

    def get_s3_json_content(self, bucket, key):
        if key == DONOR_LIST_KEY:
            return self.cts_data
        if key == LR_CURVES_SIMILARITY_TO_PROBABILITY:
            return self.lrs_data


def create_cat_test_ctx():
    ctx = Context()
    ctx['utils'] = MockCategoricalData()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    return ctx.child()


def create_cts_test_ctx():
    ctx = Context()
    ctx['utils'] = MockContinuousData()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    return ctx.child()


def test_soft_fail():
    # Create a new instance of a SimilarityRecommender.
    ctx = Context()
    ctx['utils'] = MockNoDataUtils()
    ctx['clock'] = Clock()
    ctx['cache'] = JSONCache(ctx)
    r = SimilarityRecommender(ctx)

    # Don't recommend if the source files cannot be found.
    assert not r.can_recommend({})


def test_can_recommend():
    # Create a new instance of a SimilarityRecommender.
    ctx = create_cts_test_ctx()
    r = SimilarityRecommender(ctx)

    # Test that we can't recommend if we have not enough client info.
    assert not r.can_recommend({})

    # Test that we can recommend for a normal client.
    assert r.can_recommend(generate_a_fake_taar_client())

    # Check that we can not recommend if any required client field is missing.
    required_fields = CATEGORICAL_FEATURES + CONTINUOUS_FEATURES

    for required_field in required_fields:
        profile_without_x = generate_a_fake_taar_client()

        # Make an empty value in a required field in the client info dict.
        profile_without_x[required_field] = None
        assert not r.can_recommend(profile_without_x)

        # Completely remove (in place) the entire required field from the dict.
        del profile_without_x[required_field]
        assert not r.can_recommend(profile_without_x)


def test_recommendations():
    # Create a new instance of a SimilarityRecommender.
    ctx = create_cts_test_ctx()
    r = SimilarityRecommender(ctx)

    # TODO: clobber the SimilarityRecommender::lr_curves

    recommendation_list = r.recommend(generate_a_fake_taar_client(), 1)

    assert isinstance(recommendation_list, list)
    assert len(recommendation_list) == 1

    recommendation, weight = recommendation_list[0]

    # Make sure that the reported addons are the expected ones from the most similar donor.
    assert "{test-guid-1}" == recommendation
    assert type(weight) == np.float64


def test_recommender_str():
    # Tests that the string representation of the recommender is correct.
    ctx = create_cts_test_ctx()
    r = SimilarityRecommender(ctx)
    assert str(r) == "SimilarityRecommender"


def test_get_lr():
    # Tests that the likelihood ratio values are not empty for extreme values and are realistic.
    ctx = create_cts_test_ctx()
    r = SimilarityRecommender(ctx)
    assert r.get_lr(0.0001) is not None
    assert r.get_lr(10.0) is not None
    assert r.get_lr(0.001) > r.get_lr(5.0)


def test_compute_clients_dist():
    # Test the distance function computation.
    ctx = create_cts_test_ctx()
    r = SimilarityRecommender(ctx)
    test_clients = [
        {
            "client_id": "test-client-002",
            "activeAddons": [],
            "geo_city": "sfo-us",
            "subsession_length": 1,
            "locale": "en-US",
            "os": "windows",
            "bookmark_count": 1,
            "tab_open_count": 1,
            "total_uri": 1,
            "unique_tlds": 1
        },
        {
            "client_id": "test-client-003",
            "activeAddons": [],
            "geo_city": "brasilia-br",
            "subsession_length": 1,
            "locale": "br-PT",
            "os": "windows",
            "bookmark_count": 10,
            "tab_open_count": 1,
            "total_uri": 1,
            "unique_tlds": 1
        },
        {
            "client_id": "test-client-004",
            "activeAddons": [],
            "geo_city": "brasilia-br",
            "subsession_length": 100,
            "locale": "br-PT",
            "os": "windows",
            "bookmark_count": 10,
            "tab_open_count": 10,
            "total_uri": 100,
            "unique_tlds": 10
        }
    ]
    per_client_test = []

    # Compute a different set of distances for each set of clients.
    for tc in test_clients:
        test_distances = r.compute_clients_dist(tc)
        assert len(test_distances) == len(CONTINUOUS_FEATURE_FIXTURE_DATA)
        per_client_test.append(test_distances[2][0])

    # Ensure the different clients also had different distances to a specific donor.
    assert per_client_test[0] >= per_client_test[1] >= per_client_test[2]


def test_distance_functions():
    # Tests the similarity functions via expected output when passing modified client data.
    ctx = create_cts_test_ctx()
    r = SimilarityRecommender(ctx)

    # Generate a fake client.
    test_client = generate_a_fake_taar_client()
    recs = r.recommend(test_client, 10)
    assert len(recs) > 0

    # Make it a generally poor match for the donors.
    test_client.update({'total_uri': 10, 'bookmark_count': 2, 'subsession_length': 10})

    all_client_values_zero = test_client
    # Make all categorical variables non-matching with any donor.
    all_client_values_zero.update({key: 'zero' for key in test_client.keys() if key in CATEGORICAL_FEATURES})
    recs = r.recommend(all_client_values_zero, 10)
    assert len(recs) == 0

    # Make all continuous variables equal to zero.
    all_client_values_zero.update({key: 0 for key in test_client.keys() if key in CONTINUOUS_FEATURES})
    recs = r.recommend(all_client_values_zero, 10)
    assert len(recs) == 0

    # Make all categorical variables non-matching with any donor.
    all_client_values_high = test_client
    all_client_values_high.update({key: 'one billion' for key in test_client.keys() if key in CATEGORICAL_FEATURES})
    recs = r.recommend(all_client_values_high, 10)
    assert len(recs) == 0

    # Make all continuous variables equal to a very high numerical value.
    all_client_values_high.update({key: 1e60 for key in test_client.keys() if key in CONTINUOUS_FEATURES})
    recs = r.recommend(all_client_values_high, 10)
    assert len(recs) == 0

    # Test for 0.0 values if j_c is not normalized and j_d is fine.
    j_c = 0.0
    j_d = 0.42
    assert abs(j_c * j_d) == 0.0
    assert abs((j_c + 0.01) * j_d) != 0.0


def test_weights_continuous():
    # Create a new instance of a SimilarityRecommender.
    ctx = create_cts_test_ctx()
    r = SimilarityRecommender(ctx)

    # In the ensemble method recommendations should be a sorted list of tuples
    # containing [(guid, weight), (guid, weight)... (guid, weight)].
    recommendation_list = r.recommend(generate_a_fake_taar_client(), 2)
    with open('/tmp/similarity_recommender.json', 'w') as fout:
        fout.write(json.dumps(recommendation_list))

    # Make sure the structure of the recommendations is correct and
    # that we recommended the the right addons.

    assert len(recommendation_list) == 2
    for recommendation, weight in recommendation_list:
        assert isinstance(recommendation, six.string_types)
        assert isinstance(weight, float)

    # Test that sorting is appropriate.
    rec0 = recommendation_list[0]
    rec1 = recommendation_list[1]

    rec0_weight = rec0[1]
    rec1_weight = rec1[1]

    # Duplicate presence of test-guid-1 should mean rec0_weight is double
    # rec1_weight, and both should be greater than 1.0

    assert rec0_weight > rec1_weight > 1.0


def test_weights_categorical():
    '''
    This should get :
        ["{test-guid-1}", "{test-guid-2}", "{test-guid-3}", "{test-guid-4}"],
        ["{test-guid-9}", "{test-guid-10}", "{test-guid-11}", "{test-guid-12}"]
    from the first two entries in the sample data where the geo_city
    data

    '''
    # Create a new instance of a SimilarityRecommender.
    ctx = create_cat_test_ctx()
    ctx2 = create_cts_test_ctx()
    wrapped = ctx2.wrap(ctx)
    r = SimilarityRecommender(wrapped)

    # In the ensemble method recommendations should be a sorted list of tuples
    # containing [(guid, weight), (guid, weight)... (guid, weight)].
    recommendation_list = r.recommend(generate_a_fake_taar_client(), 2)

    assert len(recommendation_list) == 2
    # Make sure the structure of the recommendations is correct and that we recommended the the right addons.
    for recommendation, weight in recommendation_list:
        assert isinstance(recommendation, six.string_types)
        assert isinstance(weight, float)

    # Test that sorting is appropriate.
    rec0 = recommendation_list[0]
    rec1 = recommendation_list[1]

    rec0_weight = rec0[1]
    rec1_weight = rec1[1]

    assert rec0_weight > rec1_weight > 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import itertools
from .base_recommender import AbstractRecommender
import threading
import time

S3_BUCKET = 'telemetry-parquet'
ENSEMBLE_WEIGHTS = 'taar/ensemble/ensemble_weight.json'

logger = logging.getLogger(__name__)


class WeightCache:
    def __init__(self, ctx):
        self._ctx = ctx
        assert 'cache' in self._ctx

        self._lock = threading.RLock()

        self._weights = None
        self._expiry = None

    def now(self):
        return time.time()

    def getWeights(self):
        with self._lock:
            now = self.now()
            if self._expiry is not None:
                if self._expiry < now:
                    # Cache is expired.
                    self._weights = None
                    # Push expiry to 5 minutes from now
                    self._expiry = now + 300

            if self._weights is None:
                tmp = self._ctx['cache'].get_s3_json_content(S3_BUCKET, ENSEMBLE_WEIGHTS)
                self._weights = tmp['ensemble_weights']

            return self._weights


class EnsembleRecommender(AbstractRecommender):
    """
    The EnsembleRecommender is a collection of recommenders where the
    results from each recommendation is amplified or dampened by a
    factor.  The aggregate results are combines and used to recommend
    addons for users.
    """
    def __init__(self, ctx):
        self._ctx = ctx

        assert 'recommender_map' in self._ctx

        self._init_from_ctx()

    def _init_from_ctx(self):
        # Copy the map of the recommenders
        self.RECOMMENDER_KEYS = ['collaborative', 'similarity', 'locale']
        self._recommender_map = self._ctx['recommender_map']
        self._weight_cache = WeightCache(self._ctx.child())

    def can_recommend(self, client_data, extra_data={}):
        """The ensemble recommender is always going to be
        available if at least one recommender is available"""
        return sum([self._recommender_map[rkey].can_recommend(client_data)
                    for rkey in self.RECOMMENDER_KEYS])

    def recommend(self, client_data, limit, extra_data={}):
        """
        Ensemble recommendations are aggregated from individual
        recommenders.  The ensemble recommender applies a weight to
        the recommendation outputs of each recommender to reorder the
        recommendations to be a better fit.

        The intuitive understanding is that the total space of
        recommended addons across all recommenders will include the
        'true' addons that should be recommended better than any
        individual recommender.  The ensemble method simply needs to
        weight each recommender appropriate so that the ordering is
        correct.
        """

        preinstalled_addon_ids = client_data.get('installed_addons', [])

        # Compute an extended limit by adding the length of
        # the list of any preinstalled addons.
        extended_limit = limit + len(preinstalled_addon_ids)

        flattened_results = []
        ensemble_weights = self._weight_cache.getWeights()

        for rkey in self.RECOMMENDER_KEYS:
            recommender = self._recommender_map[rkey]

            if recommender.can_recommend(client_data):
                raw_results = recommender.recommend(client_data,
                                                    extended_limit,
                                                    extra_data)
                reweighted_results = []
                for guid, weight in raw_results:
                    item = (guid, weight * ensemble_weights[rkey])
                    reweighted_results.append(item)
                flattened_results.extend(reweighted_results)

        # Sort the results by the GUID
        flattened_results.sort(key=lambda item: item[0])

        # group by the guid, sum up the weights for recurring GUID
        # suggestions across all recommenders
        guid_grouper = itertools.groupby(flattened_results, lambda item: item[0])

        ensemble_suggestions = []
        for (guid, guid_group) in guid_grouper:
            weight_sum = sum([v for (g, v) in guid_group])
            item = (guid, weight_sum)
            ensemble_suggestions.append(item)

        # Sort in reverse order (greatest weight to least)
        ensemble_suggestions.sort(key=lambda x: -x[1])

        filtered_ensemble_suggestions = [(guid, weight) for (guid, weight)
                                         in ensemble_suggestions
                                         if guid not in preinstalled_addon_ids]

        results = filtered_ensemble_suggestions[:limit]

        log_data = (client_data['client_id'],
                    str(ensemble_weights),
                    str([r[0] for r in results]))
        logger.info("client_id: [%s], ensemble_weight: [%s], guids: [%s]" % log_data)
        return results



import logging
from .base_recommender import AbstractRecommender

ADDON_LIST_BUCKET = 'telemetry-parquet'
ADDON_LIST_KEY = 'taar/legacy/legacy_dict.json'


logger = logging.getLogger(__name__)


class LegacyRecommender(AbstractRecommender):
    """ A recommender class that returns potential replacements for deprecated legacy addons.

    This will load a json file (periodically updated) containing suggested web extension
    alternatives to legacy addons not compatible with Fx 57.

    This recommender may provide useful recommendations when collaborative_recommender
    may not work.
    """
    def __init__(self, ctx):
        self._ctx = ctx
        assert 'cache' in self._ctx
        self._init_from_ctx()

    def _init_from_ctx(self):
        self.legacy_replacements = self._ctx['cache'].get_s3_json_content(ADDON_LIST_BUCKET,
                                                                          ADDON_LIST_KEY)
        if self.legacy_replacements is None:
            logger.error("Cannot download the JSON resource: {}".format(ADDON_LIST_KEY))

    def can_recommend(self, client_data, extra_data={}):
        # We can't recommend if we don't have our data files.
        if self.legacy_replacements is None:
            return False

        # Get active addons from the client data and pull the legacy GUIDS.
        legacy_addons = client_data.get('disabled_addons_ids', None)

        # Can't use this recommender is no addons are installed.
        if not legacy_addons or not isinstance(legacy_addons, list):
            return False

        # If no member of the installed legacy addon guids can be found in the loaded resource
        # then no recommendation is possible.
        return len(set(legacy_addons).intersection(self.legacy_replacements.keys())) > 0

    def recommend(self, client_data, limit, extra_data={}):
        legacy_addons = client_data.get('disabled_addons_ids', [])

        replacements = [self.legacy_replacements[legacy_addon]
                        for legacy_addon in legacy_addons
                        if legacy_addon in self.legacy_replacements]

        recommendations = []
        while len(recommendations) < limit and len(replacements) > 0:
            remove_list = []
            for idx, replace_list in enumerate(replacements):
                if len(replace_list) == 0:
                    remove_list.append(idx)
                    continue
                recommendations.append((replace_list.pop(0), 1))
            remove_list.reverse()
            for idx in remove_list:
                del replacements[idx]

        log_data = (client_data['client_id'])
        logger.info("legacy_recommender_triggered, "
                    "client_id: [%s]" % log_data)

        return recommendations
